from vaulboros.api_client import ApiResponse

from vaulboros.api.approle import ApproleApi
from vaulboros.api.kv import KvApi, Kv2Api
from vaulboros.api.mount import MountApi
from vaulboros.api.token import TokenApi
from vaulboros.api.base import BaseApi, VaultToken


import asyncio
import requests
import json
import datetime

from collections import defaultdict

try:
    import aiohttp
    AIOHTTP_INSTALLED = True
except:
    AIOHTTP_INSTALLED = False

class ApiRequest:

    def __init__(self, url: str, method: str, headers: dict={}, data: str=""):
        self.url = url
        self.method = method
        self.headers = headers
        self.data = data

class ApiResponse:

    def __init__(self, url: str, method: str, headers: dict, status_code: int, content: str):
        self.url = url
        self.method = method
        self.headers = headers
        self.status_code = status_code
        self.content = content

        self.request: ApiRequest = None

    def __repr__(self) -> str:
        return f"{self.status_code} - {self.url}: {self.content}"

    @classmethod
    def from_requests_response(cls, response: requests.Response):
        
        try:
            content = json.loads(response.content.decode())
        except:
            content = response.content

        return cls(
            url = response.url,
            method = response.request.method.lower(),
            headers = response.headers,
            status_code = response.status_code,
            content = content
        )
    
    @classmethod
    async def from_aiohttp_response(cls, response):
        response: aiohttp.ClientResponse = response

        headers = response.headers
        status_code = response.status
        content = await response.json()
        url = str(response.url)
        method = response.method.lower()

        return cls(
            url, method, headers, status_code, content
        )

# очень простой кеш, который будет хранить мапу METHOD[URL] <--> ApiResponse и возвращать с него ответы
class Cache:

    def __init__(self):
        self.__cache = defaultdict(dict)
        self.stats = dict(
            number_of_queries = 0
        )

    def add(self, response: ApiResponse):
        self.__cache[response.method][response.url] = response

    def get(self, request: ApiRequest) -> ApiResponse:
        if request.url in self.__cache[request.method].keys():
            self.stats['number_of_queries'] += 1
            return self.__cache[request.method][request.url]
        
    def full(self):
        return self.__cache

    def __contains__(self, item: ApiRequest):
        return item.url in self.__cache[item.method].keys()

class BaseApiClient: # класс с общими методами для sync, async

    cache = Cache()
    logger = None

    def __init__(self, common_headers: dict={}, cache_response: bool=True) -> None:
        self.common_headers = common_headers
        self.cache_response = cache_response

    @staticmethod
    def log_response(response: ApiResponse) -> None:
        if BaseApiClient.logger is not None:
            print(f"[ {datetime.datetime.now()} ] [{response.method}] <-- {response.url} - {response.status_code}")
    
    @staticmethod
    def log_request(request: ApiRequest) -> None:
        if BaseApiClient.logger is not None:
            print(f"[ {datetime.datetime.now()} ] [{request.method}] --> {request.url} - H: {request.headers} D: {request.data}")

class AsyncApiClient(BaseApiClient):

    available = AIOHTTP_INSTALLED

    def __init__(self, common_headers={}, cache_response: bool=True):

        if not AsyncApiClient.available:
            raise Exception("You are trying to use async api client but aiohttp is not installed! It's not gonna work!")

        super().__init__(common_headers, cache_response)

    async def _request(self, request: ApiRequest) -> ApiResponse:
        response: ApiResponse = None

        self.log_request(request)

        if self.cache_response:
            if request in self.cache:
                response = self.cache.get(request)

        if response is None:
            async with aiohttp.ClientSession() as session:
                async with session.request(
                        method=request.method, 
                        url=request.url, 
                        data=request.data,
                        headers=request.headers) as raw_response:
                    
                    response = await ApiResponse.from_aiohttp_response(raw_response)

            response.request = request

            if self.cache_response:
                self.cache.add(response)

        self.log_response(response)

        return response

    async def get(self, url, headers={}) -> ApiResponse:
        response = await self._request(ApiRequest(url, 'get', {**headers, **self.common_headers}))
        return response
    
    async def post(self, url, headers={}, data="") -> ApiResponse:
        response = await self._request(ApiRequest(url, 'post', {**headers, **self.common_headers}, data))
        return response
    
    async def list(self, url, headers={}) -> ApiResponse:
        response = await self._request(ApiRequest(url, 'list', {**headers, **self.common_headers}))
        return response

class SyncApiClient(BaseApiClient):

    def __init__(self, common_headers={}, cache_response: bool=True):
        super().__init__(common_headers, cache_response)

    def _request(self, request: ApiRequest) -> ApiResponse:
        response: ApiResponse = None

        self.log_request(request)

        if self.cache_response:
            if request in self.cache:
                response = self.cache.get(request)

        if response is None:

            raw_response = requests.request(
                method=request.method,
                url=request.url,
                data=request.data,
                headers=request.headers
            )

            response = ApiResponse.from_requests_response(raw_response)

            response.request = request

            if self.cache_response:
                self.cache.add(response)

        self.log_response(response)

        return response
    
    def get(self, url, headers={}) -> ApiResponse:
        response = self._request(ApiRequest(url, 'get', {**headers, **self.common_headers}))

        return response

    def post(self, url, headers={}, data="") -> ApiResponse:
        response = self._request(ApiRequest(url, 'post', {**headers, **self.common_headers}, data))

        return response

    def list(self, url, headers={}) -> ApiResponse:
        response = self._request(ApiRequest(url, 'list', {**headers, **self.common_headers}))

        return response


from vaulboros.api_client import ApiResponse
from vaulboros.api.token import VaultToken
from vaulboros.api.base import BaseApi, VaultResponse

# логин и управление approle auth engine

class ApproleApi(BaseApi):
    
    def __init__(self, vault_url, vault_token=None, vault_approle_mount_path='approle', **kwargs):
        self.approle_mount = vault_approle_mount_path
        super().__init__(vault_url, vault_token=vault_token, **kwargs)

    def __get_login_request_params(self, role_id, secret_id) -> dict:
        return dict(
            url=f"{self.vault_url}/v1/auth/{self.approle_mount}/login",
            data={
                "role_id": role_id,
                "secret_id": secret_id
            }
        )

    def __post_process_login_response(self, response: ApiResponse, persist_token=True):
        if response.status_code == 200:
            vault_token = VaultToken.from_api_response(response)

            if vault_token:
                if persist_token:
                    self.persist_api_token(vault_token)

            return vault_token

    def login(self, role_id, secret_id, **kwargs) -> VaultResponse:
        login_response: ApiResponse = self._post(**self.__get_login_request_params(role_id, secret_id))

        return VaultResponse(login_response, self.__post_process_login_response(login_response, **kwargs))
    
    @BaseApi.async_unavailable_fallback('login')
    async def a_login(self, role_id, secret_id, **kwargs) -> VaultResponse:
        login_response: ApiResponse = await self._a_post(**self.__get_login_request_params(role_id, secret_id))

        return VaultResponse(login_response, self.__post_process_login_response(login_response, **kwargs))


from vaulboros.api_client import ApiResponse
from vaulboros.api.base import BaseApi, VaultToken
    
class TokenApi(BaseApi):

    def __init__(self, vault_url, vault_token: VaultToken=None, **kwargs):
        super().__init__(vault_url=vault_url, vault_token=vault_token, vault_token_required=True)


import asyncio
import json

from vaulboros.api_client import ApiResponse
from vaulboros.api.base import BaseApi, VaultResponse
from typing import Union

def get_obj_from_args_kwargs(obj_type:Union[object,list[object]], args:list, kwargs:dict, 
                            obj_kwargs_key:Union[str,list[str]]=None, obj_args_index:int=1):
    if type(obj_type) is not list:
        obj_type = [obj_type]

    if type(obj_kwargs_key) is not list:
        obj_kwargs_key = list(obj_kwargs_key)
    
    target_kwarg_key = None
    for obj_key in obj_kwargs_key:
        if obj_key in kwargs.keys():
            target_kwarg_key = obj_key

    if target_kwarg_key != None:
        obj = kwargs[target_kwarg_key]
    else:
        obj = args[obj_args_index]

    if type(obj) in obj_type:
        return obj, target_kwarg_key
    else:
        return False, None

class KvSecret: # контейнер для секрета
    
    def __init__(self, secret_path):
        self.path = secret_path
        self.path_parts = list(filter(bool,self.path.split('/')))
        self.inner_path = '/'.join(self.path_parts[1:]) # путь без secret engine
        self.location_path = '/'.join(self.path_parts[:-1])
        
        self.name = self.path_parts[-1]
        self.secret_engine_name = self.path_parts[0]

        self.data: dict = dict()
        self.version = None # версия актуальна только в контексте kv v2

        self.api_response: ApiResponse = None

        #--conditions--
        self.not_found = False

    @property
    def json(self):
        return json.dumps(self.dict)

    @property
    def dict(self):
        return self.data

    def __repr__(self) -> str:
        return self.path

    @classmethod
    def from_string_path(cls, path:str):
        return cls(
            secret_path = path
        )
    
    def update_from_get_kv1_api_response(self, response: ApiResponse):
        if response.status_code == 200:
            self.data = response.content['data']

        self.api_response = response

    def update_from_get_kv2_api_response(self, response: ApiResponse):
        if response.status_code == 200:
            self.data = response.content['data']['data']

        self.api_response = response

    @classmethod
    def from_kv1_api_response(cls, response: ApiResponse):
        pass

    @classmethod
    def from_kv2_api_response(cls, response: ApiResponse):
        pass

class KvLocation: # контейнер для локейшена в kv secret engine. Содержит KvSecret и другие KvLocation

    def __init__(self, location_path=None):
        self.__path = location_path

        self.__set_paths()

        self.secrets: list[KvSecret] = list()
        self.locations: list[KvLocation] = list()

        self.api_response: ApiResponse = None

        self.status = dict()

    @property
    def path(self):
        return self.__path
    
    @path.setter
    def path(self, new_path_value):
        self.__path = new_path_value
        self.__set_paths()

    def __set_paths(self):
        if self.__path != None:
            self.path_parts = list(filter(bool,self.path.split('/')))
            self.inner_path = '/'.join(self.path_parts[1:]) # путь без secret engine
            
            self.name = self.path_parts[-1]
            self.secret_engine_name = self.path_parts[0]

    @property
    def dict(self):
        result_dict = dict()

        for secret in self.secrets:
            result_dict[secret.name] = secret.dict

        for location in self.locations:
            result_dict[location.name] = location.dict

        return result_dict

    @property
    def json(self):
        return json.dumps(self.dict)
    
    @property
    def number_of_secrets(self):
        secrets_num = len(self.secrets)
        for location in self.locations:
            secrets_num += location.number_of_secrets

        return secrets_num
    
    def __contains__(self, secret_obj:Union[str,KvSecret]):
        s_path = secret_obj if type(secret_obj) is str else secret_obj.path
        return s_path in list(map(lambda s: s.path, self.secrets))
    
    def __repr__(self) -> str:
        return self.path

    def update_from_list_api_response(self, response: ApiResponse): # kv and kv2 совместимое API
        if response.status_code == 200:
            for key in response.content['data']['keys']:
                if key.endswith('/'):
                    self.locations.append(KvLocation('/'.join(self.path_parts+[key])))
                else:
                    self.secrets.append(KvSecret('/'.join(self.path_parts+[key])))

        self.api_response = response

    @classmethod
    def from_string_path(cls, path:str):
        return cls(
            location_path = path
        )

class KvApi(BaseApi):

    def __init__(self, vault_url, vault_kv_engine_path, vault_token=None, **kwargs):
        super().__init__(vault_url=vault_url, vault_token=vault_token, vault_token_required=True, **kwargs)

        self.path = vault_kv_engine_path
        self.name = list(filter(bool,self.path.split('/')))[0]

    def secret_engine_obj_unification(unify_type: Union[KvSecret,KvLocation]):
        def obj_unification_wrapper(func):
            def wrapper(*args, **kwargs):
                args = list(args)

                target_obj, target_obj_kwargs_key = get_obj_from_args_kwargs([KvSecret,KvLocation,str], args, kwargs, ['secret_obj','location_obj'])

                if type(target_obj) == str:
                    target_obj = unify_type.from_string_path(target_obj)

                if target_obj_kwargs_key:
                    kwargs[target_obj_kwargs_key] = target_obj
                else:
                    args[1] = target_obj

                if asyncio.iscoroutinefunction(func):
                    async def awrapper():
                        return await func(*args, **kwargs)
                    return awrapper()
                else:
                    return func(*args, **kwargs)
            
            return wrapper
        return obj_unification_wrapper

    def check_secret_engine_obj_path(func):
        def wrapper(*args, **kwargs):
            target_obj, _ = get_obj_from_args_kwargs([KvSecret,KvLocation], args, kwargs, ['secret_obj','location_obj'])

            api_obj: KvApi = args[0]

            if api_obj.name != target_obj.secret_engine_name:
                raise Exception(f"Trying to process {type(target_obj)} for '{target_obj.secret_engine_name}' secret engine in '{api_obj.name}' secret engine api object")
            else:
                if asyncio.iscoroutinefunction(func):
                    async def awrapper():
                        return await func(*args, **kwargs)
                    return awrapper()
                else:
                    return func(*args, **kwargs)

        return wrapper
    
    def _get_get_secret_request_params(self, secret_obj:KvSecret) -> dict:

        return dict(
            url=f"{self.vault_url}/v1/{self.name}/{secret_obj.inner_path}"
        )
    
    def _post_process_get_secret_response(self, secret_obj:KvSecret, response: ApiResponse) -> KvSecret:
        if response.status_code == 200:
            secret_obj.update_from_get_kv1_api_response(response)
        
        return secret_obj

    @secret_engine_obj_unification(KvSecret)
    @check_secret_engine_obj_path
    def get_secret(self, secret_obj:Union[str,KvSecret]) -> VaultResponse:
        secret_response: ApiResponse = self._get(**self._get_get_secret_request_params(secret_obj))
        return VaultResponse(secret_response, self._post_process_get_secret_response(secret_obj, secret_response))
    
    @BaseApi.async_unavailable_fallback('get_secret')
    @secret_engine_obj_unification(KvSecret)
    @check_secret_engine_obj_path
    async def a_get_secret(self, secret_obj:Union[str,KvSecret]) -> VaultResponse:
        secret_response = await self._a_get(**self._get_get_secret_request_params(secret_obj))
        return VaultResponse(secret_response, self._post_process_get_secret_response(secret_obj, secret_response))

    def _get_list_secrets_request_params(self, location_obj:KvLocation) -> dict:
        return dict(
            url=f"{self.vault_url}/v1/{self.name}/{location_obj.inner_path}"
        )

    def _post_process_list_location_response(self, location_obj:KvLocation, response: ApiResponse) -> KvLocation:
        location_obj.update_from_list_api_response(response)

        return location_obj

    @secret_engine_obj_unification(KvLocation)
    @check_secret_engine_obj_path
    def list_location(self, location_obj:Union[str,KvLocation]):
        location_reponse = self._list(**self._get_list_secrets_request_params(location_obj))
        return VaultResponse(location_reponse, self._post_process_list_location_response(location_obj, location_reponse))

    @BaseApi.async_unavailable_fallback('list_location')
    @secret_engine_obj_unification(KvLocation)
    @check_secret_engine_obj_path
    async def a_list_location(self, location_obj:Union[str,KvLocation]):
        location_response = await self._a_list(**self._get_list_secrets_request_params(location_obj))
        return VaultResponse(location_response, self._post_process_list_location_response(location_obj, location_response))
    
    @secret_engine_obj_unification(KvLocation)
    @check_secret_engine_obj_path
    def load_location(self, location_obj:Union[str,KvLocation], recursive=False):
        if location_obj.api_response is None: # значит ничего для локейшена не спрашивалось
            self.list_location(location_obj)

        for secret in location_obj.secrets:
            self.get_secret(secret)

        if recursive:
            for location in location_obj.locations:
                self.load_location(location, recursive=True)

        return location_obj
    
    
    @BaseApi.async_unavailable_fallback('load_location')
    @secret_engine_obj_unification(KvLocation)
    @check_secret_engine_obj_path
    async def a_load_location(self, location_obj:Union[str,KvLocation], recursive=False, event_loop:asyncio.AbstractEventLoop=None):

        if event_loop is None:
            event_loop = asyncio.get_running_loop()

        if location_obj.api_response is None: # значит ничего для локейшена не спрашивалось
            #await event_loop.create_task(self.a_list_location(location_obj))
            await self.a_list_location(location_obj)

        tasks: list[asyncio.Task] = []

        if recursive:
            for location in location_obj.locations:
                tasks.append(event_loop.create_task(self.a_load_location(location, recursive=True, event_loop=event_loop)))

        for secret in location_obj.secrets:
            tasks.append(event_loop.create_task(self.a_get_secret(secret)))

        if len(tasks) != 0:
            await asyncio.wait(tasks)

        return location_obj

    def __get_create_secret_request_params(self):
        pass

    def create_secret(self):
        pass

    async def a_create_secret(self):
        pass

    def update_secret(self):
        self.create_secret()

    async def a_update_secret(self):
        await self.a_create_secret()

    def __get_delete_secret_request_params(self):
        pass

    def delete_secret(self):
        pass

    async def delete_secret(self):
        pass

class Kv2Api(KvApi):

    def __init__(self, vault_url, vault_kv_engine_path, vault_token=None, **kwargs):
        super().__init__(vault_url=vault_url, vault_token=vault_token, vault_kv_engine_path=vault_kv_engine_path,**kwargs)

    def _get_get_secret_request_params(self, secret_obj:KvSecret) -> dict:
    
        return dict(
            url=f"{self.vault_url}/v1/{self.name}/data/{secret_obj.inner_path}"
        )
    
    def _get_list_secrets_request_params(self, location_obj:KvLocation) -> dict:
        return dict(
            url=f"{self.vault_url}/v1/{self.name}/metadata/{location_obj.inner_path}"
        )
    
    def _post_process_get_secret_response(self, secret_obj: KvSecret, response: ApiResponse) -> KvSecret:
        secret_obj.update_from_get_kv2_api_response(response)

        return secret_obj 


import inspect
from vaulboros.api_client import SyncApiClient, AsyncApiClient, ApiResponse

class VaultResponse:

    def __init__(self, api_reponse: ApiResponse, response_entity: any=None):
        self.result = response_entity
        self.api_response = api_reponse

class VaultToken:
    
    def __init__(self, vault_token):
        self.token = vault_token

    def __repr__(self) -> str:
        return str(self.token)

    @classmethod
    def from_api_response(cls, response: ApiResponse):
        token = cls(vault_token = response.content['auth']['client_token'])
        return token 

class Headers:

    def __init__(self, sync_client: SyncApiClient, async_client: AsyncApiClient):
        self.sync_client_headers = sync_client.common_headers
        
        self.async_client_headers = False if not async_client else async_client.common_headers

    def __setitem__(self, key, value):
        header_value = str(value)
        self.sync_client_headers[key] = header_value

        if self.async_client_headers is not False:
            self.async_client_headers[key] = header_value

    def __getitem__(self, key):
        if key in self.sync_client_headers.keys():
            return_header = self.sync_client_headers[key]
        else:
            return_header = None
        return return_header
    
    # def __repr__(self) -> str:
    #     return str(self.sync_client_headers)

class BaseApi: # базовый класс для API
        

    def __init__(self, vault_url=None, vault_token: VaultToken=None, vault_token_required=False,
                    sync_client: SyncApiClient=None, async_client: AsyncApiClient=None, headers=None):

        self.vault_url = vault_url
        self.__sync_client = sync_client if sync_client is not None else SyncApiClient()
        self.__async_client = async_client if async_client is not None else AsyncApiClient() if AsyncApiClient.available else False

        self.headers = headers if headers is not None else Headers(self.__sync_client, self.__async_client)

        if self.vault_token is None and vault_token is not None: # если хедер уже не установлен, но токен передается в конструктор - устанавливаем
            self.vault_token = vault_token

        if vault_token_required and self.vault_token is None:
            raise Exception(f"No vault token provided via parameter or via another Api object. Can't create {__class__.__name__} object")
        

    @staticmethod
    def async_unavailable_fallback(fallback_func_name): # если async не поддерживается, то будет вызван синхронный метод (и разницы с запуском в синхронном режиме не будет)
        def fallback_decorator(target_func):
            async def wrapper(self, *args, **kwargs):
                if not self.__async_client:
                    return getattr(self,fallback_func_name)(*args, **kwargs) # это синхронный метод
                else:
                    return await target_func(self, *args, **kwargs) # это асинхронный метод
            return wrapper
        return fallback_decorator

    def persist_api_token(self, token: VaultToken):
        self.vault_token = token

    @property
    def vault_token(self):
        return self.headers['X-Vault-Token']
    
    @vault_token.setter
    def vault_token(self, token):
        self.headers['X-Vault-Token'] = token

    def _get(self, *args, **kwargs) -> ApiResponse:
        return self.__sync_client.get(*args, **kwargs)

    def _post(self, *args, **kwargs) -> ApiResponse:
        return self.__sync_client.post(*args, **kwargs)

    def _list(self, *args, **kwargs) -> ApiResponse:
        return self.__sync_client.list(*args, **kwargs)

    async def _a_get(self, *args, **kwargs) -> ApiResponse:
        return await self.__async_client.get(*args, **kwargs)

    async def _a_post(self, *args, **kwargs) -> ApiResponse:
        return await self.__async_client.post(*args, **kwargs)

    async def _a_list(self, *args, **kwargs) -> ApiResponse:
        return await self.__async_client.list(*args, **kwargs)

    @classmethod
    def from_api_instance(cls, api_instance, **kwargs): # создать API объект, копируя некоторые базовые сущности из другого (url, api_client)
        return cls(
            vault_url = api_instance.vault_url,
            sync_client = api_instance._BaseApi__sync_client, # private method name mangling
            async_client = api_instance._BaseApi__async_client,
            headers = api_instance.headers,
            **kwargs
        )